# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from spinn_utilities.abstract_base import abstractmethod, AbstractBase
if TYPE_CHECKING:
    from spinnman.messages.eieio.command_messages import EIEIOCommandHeader
    from spinnman.messages.eieio.data_messages import EIEIODataHeader


class AbstractEIEIOMessage(object, metaclass=AbstractBase):
    """
    Interface for an EIEIOMessage.
    """
    __slots__ = ()

    @property
    @abstractmethod
    def eieio_header(self) -> Union[EIEIOCommandHeader, EIEIODataHeader]:
        """
        The header of the message.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def bytestring(self) -> bytes:
        """
        The bytes of the message.

        :rtype: bytes
        """
        raise NotImplementedError
