# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)
from .application_copy_run_process import ApplicationCopyRunProcess
from .application_run_process import ApplicationRunProcess

from .fixed_connection_selector import FixedConnectionSelector
from .get_heap_process import GetHeapProcess
from .get_cpu_info_process import GetCPUInfoProcess
from .get_exclude_cpu_info_process import GetExcludeCPUInfoProcess
from .get_include_cpu_info_process import GetIncludeCPUInfoProcess
from .get_machine_process import GetMachineProcess
from .get_n_cores_in_state_process import GetNCoresInStateProcess
from .get_routes_process import GetMultiCastRoutesProcess
from .get_tags_process import GetTagsProcess
from .get_version_process import GetVersionProcess
from .load_routes_process import LoadMultiCastRoutesProcess
from .malloc_sdram_process import MallocSDRAMProcess
from .most_direct_connection_selector import MostDirectConnectionSelector
from .read_fixed_route_routing_entry_process import (
    ReadFixedRouteRoutingEntryProcess)
from .load_fixed_route_routing_entry_process import (
    LoadFixedRouteRoutingEntryProcess)
from .read_iobuf_process import ReadIOBufProcess
from .read_memory_process import ReadMemoryProcess
from .read_router_diagnostics_process import ReadRouterDiagnosticsProcess
from .round_robin_connection_selector import RoundRobinConnectionSelector
from .send_single_command_process import SendSingleCommandProcess
from .write_memory_process import WriteMemoryProcess

# Old name
AbstractMultiConnectionProcessConnectionSelector = ConnectionSelector

__all__ = ["AbstractMultiConnectionProcessConnectionSelector",
           "ConnectionSelector",
           "FixedConnectionSelector", "MostDirectConnectionSelector",
           "RoundRobinConnectionSelector",
           "AbstractMultiConnectionProcess",
           "ApplicationRunProcess", "ApplicationCopyRunProcess",
           "GetCPUInfoProcess",
           "GetExcludeCPUInfoProcess", "GetIncludeCPUInfoProcess",
           "GetHeapProcess",
           "GetMachineProcess", "GetMultiCastRoutesProcess",
           "GetNCoresInStateProcess", "GetTagsProcess",
           "GetVersionProcess", "LoadFixedRouteRoutingEntryProcess",
           "LoadMultiCastRoutesProcess", "MallocSDRAMProcess",
           "ReadFixedRouteRoutingEntryProcess", "ReadIOBufProcess",
           "ReadMemoryProcess", "ReadRouterDiagnosticsProcess",
           "SendSingleCommandProcess", "WriteMemoryProcess"]
