# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import IntEnum


class ProxyProtocol(IntEnum):
    """
    Websocket binary frame type identifiers in the Spalloc Proxy protocol.
    """
    #: Message relating to opening a channel
    OPEN = 0
    #: Message relating to closing a channel
    CLOSE = 1
    #: Message sent on a channel
    MSG = 2
    #: Message relating to opening an unbound listen-only channel
    OPEN_UNBOUND = 3
    #: Message sent on an unbound channel to a given board
    MSG_TO = 4
    #: Message relating to an error when opening or closing a channel
    ERROR = 5
